import pathlib

import typer

from rbx import annotations, console
from rbx.box import code, package, remote
from rbx.box.code import SanitizationLevel
from rbx.box.sanitizers import warning_stack
from rbx.box.schema import CodeItem

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


def _compile_out():
    return package.get_build_path() / 'exe'


def _compile(item: CodeItem, sanitized: SanitizationLevel, warnings: bool):
    console.console.print(f'Compiling [item]{item.path}[/item]...')
    digest = code.compile_item(item, sanitized, force_warnings=warnings, verbose=True)
    cacher = package.get_file_cacher()
    out_path = _compile_out()
    cacher.get_file_to_path(digest, out_path)
    out_path.chmod(0o755)

    # Clear the warning stack as we don't really worry about it when running
    # `rbx compile`.
    warning_stack.get_warning_stack().clear()

    console.console.print(
        f'[success]Compiled file written at [item]{out_path}[/item][/success]'
    )


def any(path: str, sanitized: bool = False, warnings: bool = False):
    pkg = package.find_problem_package_or_die()

    path = str(remote.expand_file(path))

    solution = package.get_solution_or_nil(path)
    if solution is not None:
        _compile(
            solution,
            sanitized=SanitizationLevel.FORCE if sanitized else SanitizationLevel.NONE,
            warnings=warnings,
        )
        return

    for generator in pkg.generators:
        if generator.path == pathlib.Path(path) or generator.name == path:
            _compile(
                generator,
                sanitized=SanitizationLevel.FORCE
                if sanitized
                else SanitizationLevel.PREFER,
                warnings=warnings,
            )
            return

    if pkg.checker is not None and pkg.checker.path == pathlib.Path(path):
        _compile(
            pkg.checker,
            sanitized=SanitizationLevel.FORCE
            if sanitized
            else SanitizationLevel.PREFER,
            warnings=warnings,
        )
        return

    if pkg.validator is not None and pkg.validator.path == pathlib.Path(path):
        _compile(
            pkg.validator,
            sanitized=SanitizationLevel.FORCE
            if sanitized
            else SanitizationLevel.PREFER,
            warnings=warnings,
        )
        return

    _compile(
        CodeItem(path=pathlib.Path(path)),
        sanitized=SanitizationLevel.FORCE if sanitized else SanitizationLevel.NONE,
        warnings=warnings,
    )



from typing import List, Optional

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import Footer, Header, Label, ListItem, ListView, RichLog

from rbx.box import package
from rbx.box.schema import TaskType
from rbx.box.testcase_extractors import (
    GenerationTestcaseEntry,
    extract_generation_testcases_from_groups,
)
from rbx.box.ui.utils.run_ui import get_metadata_markup
from rbx.box.ui.widgets.file_log import FileLog
from rbx.box.ui.widgets.rich_log_box import RichLogBox
from rbx.box.ui.widgets.test_output_box import TestBoxWidget, TestcaseRenderingData


class TestExplorerScreen(Screen):
    BINDINGS = [
        ('q', 'app.pop_screen', 'Quit'),
        ('m', 'toggle_metadata', 'Toggle metadata'),
        ('1', 'show_output', 'Show output'),
        ('2', 'show_stderr', 'Show stderr'),
        ('3', 'show_log', 'Show log'),
    ]

    def __init__(self):
        super().__init__()
        self._entries: List[GenerationTestcaseEntry] = []

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Horizontal(id='test-explorer'):
            with Vertical(id='test-list-container'):
                yield ListView(id='test-list')
            with Vertical(id='test-details'):
                yield FileLog(id='test-input')
                yield TestBoxWidget(id='test-output')
                yield RichLogBox(id='test-metadata')

    async def on_mount(self):
        self.query_one('#test-list').border_title = 'Tests'
        self.query_one('#test-input').border_title = 'Input'

        # Ensure either output or interaction is visible.
        self.action_show_output()

        metadata = self.query_one('#test-metadata', RichLogBox)
        metadata.display = False
        metadata.border_title = 'Metadata'
        metadata.wrap = True
        metadata.markup = True
        metadata.clear().write('No test selected')
        await self._update_tests()

    def action_toggle_metadata(self):
        metadata = self.query_one('#test-metadata', RichLogBox)
        metadata.display = not metadata.display

    def _update_selected_test(self, index: Optional[int]):
        input = self.query_one('#test-input', FileLog)
        output = self.query_one('#test-output', TestBoxWidget)
        metadata = self.query_one('#test-metadata', RichLog)

        if index is None:
            input.path = None
            output.reset()
            metadata.clear().write('No test selected')
            return
        entry = self._entries[index]
        input.path = entry.metadata.copied_to.inputPath

        assert entry.metadata.copied_to.outputPath is not None
        output.data = TestcaseRenderingData.from_one_path(
            entry.metadata.copied_to.outputPath
        )

        metadata.clear()
        metadata.write(get_metadata_markup(entry))

    async def _update_tests(self):
        self.watch(
            self.query_one('#test-list', ListView),
            'index',
            self._update_selected_test,
        )

        self._entries = await extract_generation_testcases_from_groups()

        test_names = [
            f'{entry.group_entry.group}/{entry.group_entry.index}'
            for entry in self._entries
        ]

        await self.query_one('#test-list', ListView).clear()
        await self.query_one('#test-list', ListView).extend(
            [ListItem(Label(name)) for name in test_names]
        )

    def is_interactive(self) -> bool:
        pkg = package.find_problem_package_or_die()
        return pkg.type == TaskType.COMMUNICATION

    def action_show_output(self):
        if self.is_interactive():
            self.query_one('#test-output', TestBoxWidget).show_interaction()
        else:
            self.query_one('#test-output', TestBoxWidget).show_output()

    def action_show_stderr(self):
        self.query_one('#test-output', TestBoxWidget).show_stderr()

    def action_show_log(self):
        self.query_one('#test-output', TestBoxWidget).show_log()



import pathlib

import typer

from rbx import annotations
from rbx.box.tooling.boca.scrape import scrape_boca

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


@app.command('scrape', help='Scrape runs from BOCA.')
def scrape() -> None:
    scrape_boca(pathlib.Path())



from __future__ import generators

import collections
import dataclasses
import pathlib
import shutil
from collections.abc import Iterator
from typing import Dict, Iterable, List, Optional, Set, Tuple

import rich
import rich.live
import rich.markup
import rich.table
import rich.text
import typer
from ordered_set import OrderedSet
from pydantic import BaseModel

from rbx import console, utils
from rbx.box import checkers, code, limits_info, package, remote, state
from rbx.box.code import (
    SanitizationLevel,
    compile_item,
    find_language_name,
)
from rbx.box.deferred import Deferred
from rbx.box.environment import (
    VerificationLevel,
)
from rbx.box.formatting import get_formatted_memory, get_formatted_time, href
from rbx.box.generators import (
    GenerationMetadata,
    expand_generator_call,
    generate_output_for_testcase,
    generate_standalone,
)
from rbx.box.sanitizers import issue_stack
from rbx.box.schema import (
    ExpectedOutcome,
    GeneratorCall,
    Solution,
    TaskType,
    Testcase,
    TestcaseGroup,
)
from rbx.box.tasks import (
    get_limits_for_language,
    run_solution_on_testcase,
)
from rbx.box.testcase_extractors import extract_generation_testcases
from rbx.box.testcase_utils import (
    TestcaseEntry,
    TestcaseInteractionParsingError,
    find_built_testcases,
    parse_interaction,
    print_interaction,
)
from rbx.grading import grading_context
from rbx.grading.limits import Limits
from rbx.grading.steps import (
    Evaluation,
    Outcome,
)
from rbx.utils import StatusProgress

StructuredEvaluation = Dict[str, Dict[str, List[Optional[Deferred[Evaluation]]]]]


@dataclasses.dataclass(frozen=True)
class EvaluationItem:
    solution: Solution
    testcase_entry: TestcaseEntry
    eval: Deferred[Evaluation]


class GroupSkeleton(BaseModel):
    name: str
    testcases: List[Testcase]


class SolutionSkeleton(Solution):
    runs_dir: pathlib.Path

    def get_entry_prefix(self, entry: TestcaseEntry) -> pathlib.Path:
        return self.runs_dir / entry.group / f'{entry.index:03d}'


class SolutionReportSkeleton(BaseModel):
    solutions: List[SolutionSkeleton]
    entries: List[TestcaseEntry]
    groups: List[GroupSkeleton]
    limits: Dict[str, Limits]
    verification: VerificationLevel
    capture_pipes: bool = False

    def get_solution_limits(self, solution: Solution) -> Limits:
        lang = code.find_language_name(solution)
        if lang is None:
            return limits_info.get_package_limits(self.verification)
        return self.limits[lang]

    def get_solution_limits_from_disk(self, solution: Solution) -> Limits:
        lang = code.find_language_name(solution)
        return limits_info.get_limits(
            language=lang,
            profile=self.get_solution_limits(solution).profile,
            verification=self.verification,
        )

    def find_group_skeleton(self, group_name: str) -> Optional[GroupSkeleton]:
        groups = [group for group in self.groups if group.name == group_name]
        if not groups:
            return None
        return groups[0]

    def find_solution_skeleton(self, solution: Solution) -> Optional[SolutionSkeleton]:
        for sol in self.solutions:
            if sol.path == solution.path:
                return sol
        return None

    def find_solution_skeleton_index(self, solution: Solution) -> Optional[int]:
        for i, sol in enumerate(self.solutions):
            if sol.path == solution.path:
                return i
        return None

    def get_solution_path_set(self) -> Set[str]:
        return set(str(sol.path) for sol in self.solutions)

    def empty_structured_evaluation(self) -> StructuredEvaluation:
        res: StructuredEvaluation = {}
        for solution in self.solutions:
            res[str(solution.path)] = {}
            for group in self.groups:
                res[str(solution.path)][group.name] = [None for _ in group.testcases]
        return res


@dataclasses.dataclass
class RunSolutionResult:
    skeleton: SolutionReportSkeleton
    items: List[EvaluationItem]

    def empty_structured_evaluation(self) -> StructuredEvaluation:
        return self.skeleton.empty_structured_evaluation()


class FailedSolutionIssue(issue_stack.Issue):
    def __init__(self, solution: Solution):
        self.solution = solution

    def get_detailed_section(self) -> Tuple[str, ...]:
        return ('solutions',)

    def get_detailed_message(self) -> str:
        return f'[item]{href(self.solution.path)}[/item] has an unexpected outcome.'


def is_fast(solution: Solution) -> bool:
    # If solution has TLE tag, it is considered slow.
    return not solution.outcome.is_slow()


def get_matching_solutions(expected_outcome: ExpectedOutcome) -> List[Solution]:
    res = []
    for solution in package.get_solutions():
        if not solution.outcome.intersect(expected_outcome):
            continue
        res.append(solution)
    return res


def get_exact_matching_solutions(expected_outcome: ExpectedOutcome) -> List[Solution]:
    res = []
    for solution in package.get_solutions():
        if solution.outcome == expected_outcome:
            res.append(solution)
    return res


def compile_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
    sanitized: bool = False,
) -> Dict[pathlib.Path, str]:
    compiled_solutions = {}

    if tracked_solutions is None:
        tracked_solutions = set(str(sol.path) for sol in package.get_solutions())

    for solution in expand_solutions(list(tracked_solutions)):
        if progress:
            progress.update(f'Compiling solution {href(solution.path)}...')
        try:
            compiled_solutions[solution.path] = compile_item(
                solution,
                sanitized=SanitizationLevel.FORCE
                if sanitized
                else SanitizationLevel.NONE,
            )
        except:
            console.console.print(
                f'[error]Failed compiling solution {href(solution.path)}.[/error]'
            )
            raise

    return compiled_solutions


def _run_solution(
    solution: Solution,
    compiled_digest: str,
    checker_digest: Optional[str],
    runs_dir: pathlib.Path,
    group_name: str,
    interactor_digest: Optional[str] = None,
    progress: Optional[StatusProgress] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    timelimit_override: Optional[int] = None,
    nruns: int = 0,
) -> List[Deferred[Evaluation]]:
    group = package.get_testgroup(group_name)
    testcases = find_built_testcases(group)
    res: List[Deferred[Evaluation]] = []
    for i, testcase in enumerate(testcases):
        assert testcase.outputPath is not None
        output_path = runs_dir / group.name
        output_path.mkdir(parents=True, exist_ok=True)

        if progress:
            progress.update(
                f'Running solution {href(solution.path)} on test [item]{group.name}[/item] / [item]{i}[/item]...'
            )

        async def run_fn(i=i, testcase=testcase, output_path=output_path):
            return await run_solution_on_testcase(
                solution,
                compiled_digest,
                checker_digest,
                testcase,
                output_dir=output_path,
                interactor_digest=interactor_digest,
                testcase_index=i,
                verification=verification,
                timelimit_override=timelimit_override,
                nruns=nruns,
            )

        res.append(Deferred(run_fn))

    return res


async def convert_list_of_solution_evaluations_to_dict(
    skeleton: SolutionReportSkeleton,
    items: Iterable[EvaluationItem],
) -> List[Dict[str, List[Evaluation]]]:
    res: List[Dict[str, List[Evaluation]]] = [
        collections.defaultdict(list) for _ in package.get_solutions()
    ]

    for item in items:
        sol_idx = skeleton.find_solution_skeleton_index(item.solution)
        if sol_idx is not None:
            to_append = await item.eval()
            res[sol_idx][item.testcase_entry.group].append(to_append)

    return res


def _get_solutions_for_skeleton(
    tracked_solutions: Optional[Iterable[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> List[Solution]:
    solutions = [
        sol
        for sol in package.get_solutions()
        if verification.value >= VerificationLevel.ALL_SOLUTIONS.value or is_fast(sol)
    ]
    if tracked_solutions is not None:
        solutions = expand_solutions(list(tracked_solutions))
    return solutions


def _get_report_skeleton(
    tracked_solutions: Optional[Iterable[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    timelimit_override: Optional[int] = None,
) -> SolutionReportSkeleton:
    pkg = package.find_problem_package_or_die()
    solutions = _get_solutions_for_skeleton(tracked_solutions, verification)

    langs = set(find_language_name(solution) for solution in solutions)
    limits = {
        lang: get_limits_for_language(lang, verification, timelimit_override)
        for lang in langs
        if lang is not None
    }

    groups = []
    for group in pkg.testcases:
        testcases = find_built_testcases(group)
        groups.append(GroupSkeleton(name=group.name, testcases=testcases))
    entries = [
        TestcaseEntry(group=group.name, index=i)
        for group in groups
        for i in range(len(group.testcases))
    ]

    # Prepare directory.
    runs_dir = package.get_problem_runs_dir()
    shutil.rmtree(str(runs_dir), ignore_errors=True)
    runs_dir.mkdir(parents=True, exist_ok=True)

    skeleton = SolutionReportSkeleton(
        solutions=[
            SolutionSkeleton(
                **solution.model_dump(),
                runs_dir=package.get_problem_runs_dir() / f'{i}',
            )
            for i, solution in enumerate(solutions)
        ],
        groups=groups,
        limits=limits,
        entries=entries,
        verification=verification,
        capture_pipes=state.STATE.debug_logs,
    )

    skeleton_file = runs_dir / 'skeleton.yml'
    skeleton_file.write_text(utils.model_to_yaml(skeleton))

    return skeleton


def _produce_solution_items(
    skeleton: SolutionReportSkeleton,
    progress: Optional[StatusProgress] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    check: bool = True,
    timelimit_override: Optional[int] = None,
    sanitized: bool = False,
    nruns: int = 0,
) -> List[EvaluationItem]:
    pkg = package.find_problem_package_or_die()

    if pkg.type == TaskType.COMMUNICATION:
        checker_digest = (
            checkers.compile_checker() if check and pkg.checker is not None else None
        )
        interactor_digest = checkers.compile_interactor()
    else:
        checker_digest = checkers.compile_checker() if check else None
        interactor_digest = None

    compiled_solutions = compile_solutions(
        progress=progress,
        tracked_solutions=skeleton.get_solution_path_set(),
        sanitized=sanitized,
    )

    def yield_items(
        solution: SolutionSkeleton, group_name: str
    ) -> List[EvaluationItem]:
        res: List[EvaluationItem] = []
        for i, eval in enumerate(
            _run_solution(
                solution,
                compiled_solutions[solution.path],
                checker_digest,
                solution.runs_dir,
                group_name,
                interactor_digest=interactor_digest,
                progress=progress,
                verification=verification,
                timelimit_override=timelimit_override,
                nruns=nruns,
            )
        ):
            res.append(
                EvaluationItem(
                    solution=solution,
                    testcase_entry=TestcaseEntry(group=group_name, index=i),
                    eval=eval,
                )
            )

        return res

    res: List[EvaluationItem] = []

    groups = pkg.testcases
    for solution in skeleton.solutions:
        for group in groups:
            res.extend(yield_items(solution, group.name))

    return res


def print_best_output(output_files: List[pathlib.Path], empty_warning: bool = False):
    for output_file in output_files:
        if not output_file.is_file():
            continue
        if output_file.suffix == '.pio':
            try:
                print_interaction(parse_interaction(output_file))
            except TestcaseInteractionParsingError:
                # Ignore parsing errors and proceed to next file.
                continue
        else:
            console.console.print(output_file.read_text())
        return
    if empty_warning:
        console.console.print('[warning]Solution produced no output.[/warning]')


def run_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Iterable[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    check: bool = True,
    timelimit_override: Optional[int] = None,
    sanitized: bool = False,
    nruns: int = 0,
) -> RunSolutionResult:
    skeleton = _get_report_skeleton(
        tracked_solutions,
        verification=verification,
        timelimit_override=timelimit_override,
    )
    result = RunSolutionResult(
        skeleton=skeleton,
        items=_produce_solution_items(
            skeleton=skeleton,
            progress=progress,
            verification=verification,
            check=check,
            timelimit_override=timelimit_override,
            sanitized=sanitized,
            nruns=nruns,
        ),
    )
    return result


async def _generate_testcase_interactively(
    progress: Optional[StatusProgress] = None,
    generator: Optional[GeneratorCall] = None,
    testcase_entry: Optional[TestcaseEntry] = None,
    check: bool = True,
    custom_output: bool = False,
    sanitized: bool = False,
    print: bool = False,
) -> Testcase:
    main_solution = package.get_main_solution()
    irun_dir = package.get_problem_iruns_dir()
    inputs_dir = irun_dir / 'inputs'
    inputs_dir.mkdir(parents=True, exist_ok=True)
    testcase = Testcase(
        inputPath=inputs_dir / '000.in',
        outputPath=(inputs_dir / '000.out') if check else None,
    )

    is_manual = False
    is_output_manual = False
    generation_metadata = None
    if generator is not None:
        generation_metadata = GenerationMetadata(
            generator_call=expand_generator_call(generator),
            copied_to=testcase,
        )
    elif testcase_entry is not None:
        extracted = await extract_generation_testcases([testcase_entry])
        if not extracted:
            console.console.print(
                f'[error]Failed searching for testcase [item]{testcase_entry}[/item].[/error]'
            )
            raise typer.Exit(1)
        generation_metadata = extracted[0].metadata
        # Replace destination with the irun testcase we're using.
        generation_metadata.copied_to = testcase
    else:
        with utils.no_progress(progress):
            input = console.multiline_prompt('Testcase input')
        testcase.inputPath.write_text(input)
        console.console.print()

        if (
            testcase.outputPath is not None
            and not testcase.outputPath.is_file()
            and (main_solution is None or custom_output)
        ):
            with utils.no_progress(progress):
                output = console.multiline_prompt('Testcase output')
                testcase.outputPath.write_text(output)
                console.console.print()
            is_output_manual = True

        generation_metadata = GenerationMetadata(
            copied_to=testcase,
        )
        is_manual = True

    # 1. Generate testcase.
    if generation_metadata is not None:
        await generate_standalone(
            generation_metadata,
            progress=progress,
            validate=True,
        )
        if testcase_entry is not None:
            console.console.print(
                f'Using input from testcase [item]{testcase_entry}[/item].'
            )
        elif generation_metadata.generator_call is not None:
            console.console.print(
                f'Using input from generator call [item]{generation_metadata.generator_call.name} {generation_metadata.generator_call.args}[/item].'
            )
        if print and not is_manual:
            console.console.print(testcase.inputPath.read_text())
        else:
            console.console.print(
                f'Input was written to [item]{utils.abspath(testcase.inputPath)}[/item]'
            )
        console.console.print()

    # 2. Generate test output from reference
    main_solution_digest = None
    if check and not (
        testcase.outputPath is not None and testcase.outputPath.is_file()
    ):
        if main_solution is None:
            console.console.print(
                '[error]Checking is enabled but no main solution or custom output was specified.[/error]'
            )
            raise typer.Exit(1)

        if progress:
            progress.update('Compiling main solution...')
        try:
            main_solution_digest = compile_item(
                main_solution,
                sanitized=SanitizationLevel.FORCE
                if sanitized
                else SanitizationLevel.NONE,
            )
        except:
            console.console.print(
                '[error]Failed compiling main solution. If you do not want to check against a main solution, run with --nocheck flag.[/error]'
            )
            raise

    if main_solution_digest is not None and not is_output_manual:
        pkg = package.find_problem_package_or_die()
        if pkg.type == TaskType.COMMUNICATION:
            interactor_digest = checkers.compile_interactor(progress)
        else:
            interactor_digest = None

        if progress:
            progress.update('Generating output for test...')
        # TODO: Add stderr path
        if main_solution is not None:
            await generate_output_for_testcase(
                main_solution,
                main_solution_digest,
                testcase,
                interactor_digest=interactor_digest,
            )

    if check and testcase.outputPath is not None and not testcase.outputPath.is_file():
        # Output was not created, throw an error.
        console.console.print(
            '[error]Checking is enabled but no output could be generated for this testcase.[/error]'
        )
        console.console.print(
            '[error]Either specify it explicitly or provide a main solution.[/error]'
        )
        raise typer.Exit(1)

    return testcase


def _run_interactive_solutions(
    testcase: Testcase,
    skeleton: SolutionReportSkeleton,
    progress: Optional[StatusProgress] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    check: bool = True,
    sanitized: bool = False,
) -> Iterator[EvaluationItem]:
    pkg = package.find_problem_package_or_die()

    if pkg.type == TaskType.COMMUNICATION:
        checker_digest = checkers.compile_checker() if check else None
        interactor_digest = checkers.compile_interactor()
    else:
        checker_digest = checkers.compile_checker() if check else None
        interactor_digest = None

    compiled_solutions = compile_solutions(
        progress=progress,
        tracked_solutions=skeleton.get_solution_path_set(),
        sanitized=sanitized,
    )

    if progress:
        progress.update('Running solutions...')

    for solution in skeleton.solutions:
        output_dir = solution.runs_dir

        async def run_fn(solution=solution, output_dir=output_dir):
            return await run_solution_on_testcase(
                solution,
                compiled_solutions[solution.path],
                checker_digest,
                testcase,
                output_dir=output_dir,
                interactor_digest=interactor_digest,
                verification=verification,
                capture_pipes=True,
            )

        yield EvaluationItem(
            solution=solution,
            testcase_entry=TestcaseEntry(group='irun', index=0),
            eval=Deferred(run_fn),
        )


def _get_interactive_skeleton(
    tracked_solutions: Optional[Iterable[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> SolutionReportSkeleton:
    solutions = _get_solutions_for_skeleton(tracked_solutions, verification)

    langs = set(find_language_name(solution) for solution in solutions)
    limits = {
        lang: get_limits_for_language(lang, verification, timelimit_override=None)
        for lang in langs
        if lang is not None
    }

    # Ensure path is new.
    irun_dir = package.get_problem_iruns_dir()
    shutil.rmtree(str(irun_dir), ignore_errors=True)
    irun_dir.mkdir(parents=True, exist_ok=True)

    skeleton = SolutionReportSkeleton(
        solutions=[
            SolutionSkeleton(
                **solution.model_dump(),
                runs_dir=irun_dir / f'{i}',
            )
            for i, solution in enumerate(solutions)
        ],
        groups=[],
        limits=limits,
        entries=[],
        verification=verification,
        capture_pipes=True,
    )

    skeleton_file = irun_dir / 'skeleton.yml'
    skeleton_file.write_text(utils.model_to_yaml(skeleton))

    return skeleton


async def run_and_print_interactive_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Iterable[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    generator: Optional[GeneratorCall] = None,
    testcase_entry: Optional[TestcaseEntry] = None,
    check: bool = True,
    custom_output: bool = False,
    print: bool = False,
    sanitized: bool = False,
):
    pkg = package.find_problem_package_or_die()
    skeleton = _get_interactive_skeleton(
        tracked_solutions,
        verification=verification,
    )

    should_cache = testcase_entry is not None
    with grading_context.cache_level(
        grading_context.CacheLevel.CACHE_COMPILATION, when=not should_cache
    ):
        testcase = await _generate_testcase_interactively(
            progress=progress,
            generator=generator,
            testcase_entry=testcase_entry,
            check=check,
            custom_output=custom_output,
            sanitized=sanitized,
            print=print,
        )
        items = _run_interactive_solutions(
            testcase,
            skeleton=skeleton,
            progress=progress,
            verification=verification,
            check=check,
            sanitized=sanitized,
        )

    for item in items:
        sol = skeleton.find_solution_skeleton(item.solution)
        assert sol is not None

        if progress:
            progress.update(f'Running [item]{sol.path}[/item]...')

        eval = await item.eval()

        with utils.no_progress(progress):
            console.console.print(get_testcase_markup_verdict(eval), end=' ')
            _print_solution_header(sol, console.console)
            _print_solution_outcome(
                sol, skeleton, [eval], console.console, verification, subset=True
            )

        stdout_path = eval.log.stdout_absolute_path
        if print and stdout_path is not None:
            if pkg.type == TaskType.COMMUNICATION:
                console.console.rule('Interaction', style='status')
                output_files = [
                    stdout_path.with_suffix('.pio'),
                    stdout_path.with_suffix('.pout'),
                ]
                print_best_output(output_files, empty_warning=True)

            console.console.rule('Output', style='status')
            output_files = [stdout_path]
            print_best_output(output_files, empty_warning=True)
        elif stdout_path is not None:
            if stdout_path.with_suffix('.pout').is_file():
                stdout_path = stdout_path.with_suffix('.pout')

            if stdout_path.is_file():
                console.console.print(f'[status]Output:[/status] {stdout_path}')
            if stdout_path.with_suffix('.pio').is_file():
                console.console.print(
                    f'[status]Interaction:[/status] {stdout_path.with_suffix(".pio")}'
                )
            if eval.log.stderr_absolute_path is not None:
                console.console.print(
                    f'[status]Stderr:[/status] {eval.log.stderr_absolute_path}'
                )
            console.console.print()


def _get_solution_repr(sol: Solution) -> List[Tuple[str, str]]:
    fg_color = sol.outcome.style()
    return [
        ('', f'{str(sol.path)} '),
        (f'fg:{fg_color}', sol.outcome.name),
    ]


def expand_solutions_with_source(sols: List[str]) -> List[Tuple[Solution, bool]]:
    pkg_sols = {str(sol.path): sol for sol in package.get_solutions()}

    # Download remote sols.
    path_sols = remote.expand_files(sols)

    # Ensure sols exist.
    for sol in path_sols:
        if not sol.is_file():
            console.console.print(
                f'[error]Solution [item]{sol}[/item] could not be found.[/error]'
            )
            raise typer.Exit(1)

    seen_sols = set()
    res: List[Tuple[Solution, bool]] = []
    for sol in path_sols:
        if str(sol) in seen_sols:
            # This solution was already added.
            continue
        if str(sol) in pkg_sols:
            # This solution is in the package.
            res.append((pkg_sols[str(sol)], False))
        else:
            # This solution is fetched from some source.
            res.append((Solution(path=sol, outcome=ExpectedOutcome.ANY), True))
        seen_sols.add(str(sol))
    return res


def expand_solutions(sols: List[str]) -> List[Solution]:
    return [sol for sol, _ in expand_solutions_with_source(sols)]


async def pick_solutions(
    tracked_solutions: Optional[OrderedSet[str]],
    extra_solutions: Optional[List[str]] = None,
) -> List[str]:
    # Store in a separate list to maintain order with the package declaration.
    import questionary

    solutions = package.get_solutions()

    choices = [
        questionary.Choice(
            title=_get_solution_repr(sol),
            value=str(sol.path),
            checked=tracked_solutions is None or str(sol.path) in tracked_solutions,
        )
        for sol in solutions
    ]

    seen_sols = set(str(sol.path) for sol in solutions)

    if extra_solutions is not None:
        # Add only new solutions.
        choices.extend(
            questionary.Choice(
                title=_get_solution_repr(sol),
                value=str(sol.path),
                checked=True,
            )
            for sol in expand_solutions(extra_solutions)
            if str(sol.path) not in seen_sols
        )

    picked = await questionary.checkbox('Select solutions', choices=choices).ask_async()
    if picked is None:
        raise typer.Abort()
    return picked


def get_outcome_style_verdict(outcome: Outcome) -> str:
    if outcome == Outcome.ACCEPTED:
        return 'green'
    if outcome == Outcome.WRONG_ANSWER:
        return 'red'
    if outcome.is_slow():
        return 'yellow'
    if outcome == Outcome.RUNTIME_ERROR:
        return 'blue'
    if outcome == Outcome.MEMORY_LIMIT_EXCEEDED:
        return 'yellow'
    return 'magenta'


def get_outcome_markup_verdict(outcome: Outcome) -> str:
    res = '✓'
    if outcome != Outcome.ACCEPTED:
        res = '✗'
    if outcome.is_slow():
        res = '⧖'
    if outcome == Outcome.RUNTIME_ERROR:
        res = '✗'
    style = get_outcome_style_verdict(outcome)
    res = f'[{style}]{res}[/{style}]'
    return res


def get_full_outcome_markup_verdict(outcome: Outcome) -> str:
    style = get_outcome_style_verdict(outcome)
    res = f'[{style}]{outcome.name}[/{style}]'
    return res


def get_testcase_markup_verdict(eval: Evaluation) -> str:
    # if eval.log.stdout_absolute_path:
    #     output_path = eval.log.stdout_absolute_path.resolve()
    #     output_link = f'file://{output_path}'
    #     res = f'[link={output_link}]{res}[/link]'
    return get_outcome_markup_verdict(eval.result.outcome)


def get_full_testcase_markup_verdict(eval: Evaluation) -> str:
    return get_full_outcome_markup_verdict(eval.result.outcome)


def _get_evals_time_in_ms(evals: List[Evaluation]) -> int:
    if not evals:
        return 0
    evals_with_ile = [
        eval for eval in evals if eval.result.outcome == Outcome.IDLENESS_LIMIT_EXCEEDED
    ]
    for eval in evals_with_ile:
        # Try every way of estimating a ILE max timelimit.
        if eval.log.metadata is None:
            continue
        if eval.log.metadata.limits is not None:
            expanded_tl = eval.log.metadata.limits.get_expanded_tl()
            if expanded_tl is not None:
                return expanded_tl
        if eval.log.metadata.timeLimit is not None:
            return eval.log.metadata.timeLimit
    return max(int((eval.log.time or 0.0) * 1000) for eval in evals)


def _get_evals_memory_in_bytes(evals: List[Evaluation]) -> int:
    if not evals:
        return 0
    return max(int(eval.log.memory or 0) for eval in evals)


def get_evals_formatted_time(evals: List[Evaluation]) -> str:
    max_time = _get_evals_time_in_ms(evals)
    return get_formatted_time(max_time)


def get_capped_evals_formatted_time(
    limits: Limits,
    evals: List[Evaluation],
    verification: VerificationLevel,
) -> str:
    max_time = _get_evals_time_in_ms(evals)
    has_tle = any(eval.result.outcome.is_slow() for eval in evals)
    has_ile = any(
        eval.result.outcome == Outcome.IDLENESS_LIMIT_EXCEEDED for eval in evals
    )
    timelimits = [
        eval.log.metadata.limits.get_expanded_tl()
        for eval in evals
        if eval.log.metadata is not None
    ]
    timelimits = [tl for tl in timelimits if tl is not None]

    tl = None
    if timelimits:
        tl = min(timelimits)
    if tl is None:
        tl = limits.time

        if tl is not None and verification.value >= VerificationLevel.FULL.value:
            # Using double TL for verification.
            tl = tl * 2

    if tl is not None and has_tle and max_time >= tl or has_ile:
        return f'>{tl} ms'
    return f'{max_time} ms'


def get_evals_formatted_memory(evals: List[Evaluation]) -> str:
    max_memory = _get_evals_memory_in_bytes(evals)
    return get_formatted_memory(max_memory)


def get_worst_outcome(evals: List[Evaluation]) -> Outcome:
    return Outcome.worst_outcome(eval.result.outcome for eval in evals)


def get_truncated_message(message: str, max_length: int = 100) -> str:
    if len(message) > max_length:
        return message[:max_length] + '... (truncated)'
    return message


class SolutionOutcomeReport(BaseModel):
    solution: Solution
    limits: Limits
    evals: List[Evaluation]
    ok: bool
    message: Optional[Tuple[TestcaseEntry, str]]
    expectedOutcome: Optional[ExpectedOutcome]
    gotVerdicts: Set[Outcome]
    runUnderDoubleTl: bool
    doubleTlVerdicts: Set[Outcome]
    sanitizerWarnings: bool
    verification: VerificationLevel

    def get_verdict_markup(self, incomplete: bool = False) -> str:
        success_str = '[bold green]OK[/bold green]'
        if not self.ok:
            success_str = '[bold red]FAILED[/bold red]'
        if incomplete:
            success_str = '[bold yellow]INCOMPLETE[/bold yellow]'

        gotVerdicts = self.gotVerdicts if not incomplete else {}

        got_verdict_names = ' '.join(v.name for v in self.gotVerdicts)
        verdict_str = ''
        if self.expectedOutcome is not None:
            verdict_str = f'Expected: {self.expectedOutcome}'
            if gotVerdicts:
                verdict_str += f', got: {got_verdict_names}'
        elif gotVerdicts:
            verdict_str = f'Got: {got_verdict_names}'
        return f'{success_str} {verdict_str}'

    def get_verdict_markup_with_warnings(self) -> str:
        res = self.get_verdict_markup()
        if self.runUnderDoubleTl:
            if self.doubleTlVerdicts:
                res += f'\n[bold yellow]WARNING[/bold yellow] The solution still passed in double TL, but failed with [item]{" ".join(v.name for v in self.doubleTlVerdicts)}[/item].'
            else:
                res += '\n[bold yellow]WARNING[/bold yellow] The solution still passed in double TL.'
        if self.sanitizerWarnings:
            res += '\n[bold yellow]WARNING[/bold yellow] The solution had sanitizer errors or warnings, marked with [bold yellow]*[/bold yellow]. See their stderr for more details.'
        return res

    def get_outcome_markup(self, print_message: bool = True) -> str:
        res = self.get_verdict_markup_with_warnings()
        res += f'\nTime: {get_capped_evals_formatted_time(self.limits, self.evals, self.verification)}'
        res += f'\nMemory: {get_evals_formatted_memory(self.evals)}'
        if print_message and self.message is not None:
            tc, msg = self.message
            if msg:
                msg = get_truncated_message(msg)
                res += f'\nMessage for {utils.escape_markup(str(tc))}: {utils.escape_markup(msg)}'
        return res


def get_solution_outcome_report(
    solution: Solution,
    skeleton: SolutionReportSkeleton,
    evals: List[Evaluation],
    verification: VerificationLevel = VerificationLevel.NONE,
    subset: bool = False,
) -> SolutionOutcomeReport:
    has_plain_tle = False
    all_verdicts = set()
    bad_verdicts = set()
    no_tle_bad_verdicts = set()
    has_sanitizer_warnings = False
    message: Optional[Tuple[TestcaseEntry, str]] = None
    for eval, entry in zip(evals, skeleton.entries):
        all_verdicts.add(eval.result.outcome)
        if eval.result.outcome != Outcome.ACCEPTED:
            bad_verdicts.add(eval.result.outcome)
        if (
            eval.result.no_tle_outcome is not None
            and eval.result.no_tle_outcome != Outcome.ACCEPTED
        ):
            no_tle_bad_verdicts.add(eval.result.no_tle_outcome)
        has_plain_tle = has_plain_tle or (
            eval.result.outcome.is_slow() and eval.result.no_tle_outcome is None
        )
        has_sanitizer_warnings = (
            has_sanitizer_warnings or eval.result.sanitizer_warnings
        )
        if (
            eval.result.outcome
            in [
                Outcome.WRONG_ANSWER,
                Outcome.JUDGE_FAILED,
            ]
            and message is None
        ):
            message = (entry, eval.result.message)

    unmatched_bad_verdicts = set(
        v for v in bad_verdicts if not solution.outcome.match(v)
    )
    matched_bad_verdicts = bad_verdicts - unmatched_bad_verdicts
    expected_outcome_is_bad = not solution.outcome.match(Outcome.ACCEPTED)

    has_failed = unmatched_bad_verdicts or (
        expected_outcome_is_bad and not matched_bad_verdicts and not subset
    )

    report_expected_outcome = None
    report_got_verdicts = set()
    report_run_under_double_tl = False
    report_double_tl_verdicts = set()
    report_sanitizer_warnings = False
    if has_failed or not subset:
        report_expected_outcome = solution.outcome
    elif subset:
        report_got_verdicts = all_verdicts

    if has_failed or not subset:
        # Only print verdicts if not subset.
        if unmatched_bad_verdicts:
            report_got_verdicts = unmatched_bad_verdicts
        elif expected_outcome_is_bad and not matched_bad_verdicts and not subset:
            report_got_verdicts = {Outcome.ACCEPTED}

    evals_time = _get_evals_time_in_ms(evals)
    expected_outcome_is_tle = solution.outcome.matches_tle_and_is_incorrect()
    limits = skeleton.get_solution_limits(solution)
    if (
        # Running verification with double TL.
        verification.value >= VerificationLevel.FULL.value
        # Solution expects a TLE.
        and expected_outcome_is_tle
        # Solution does not have a plain TLE.
        and not has_plain_tle
        # A TLE has happened.
        and Outcome.TIME_LIMIT_EXCEEDED in matched_bad_verdicts
        # The solution runs in double TL.
        and limits.time is not None
        and evals_time < limits.time * 2
    ):
        other_verdicts = (bad_verdicts | no_tle_bad_verdicts) - {
            Outcome.TIME_LIMIT_EXCEEDED
        }
        if not other_verdicts:
            # The solution has no other bad verdicts except for TLEs in double TL.
            report_run_under_double_tl = True
        elif not (bad_verdicts - {Outcome.TIME_LIMIT_EXCEEDED}):
            # The solution has other bad soft TLE outcomes.
            report_double_tl_verdicts = other_verdicts

    if has_sanitizer_warnings:
        report_sanitizer_warnings = True

    return SolutionOutcomeReport(
        solution=solution,
        limits=skeleton.get_solution_limits(solution),
        evals=evals,
        ok=not has_failed,
        message=message,
        expectedOutcome=report_expected_outcome,
        gotVerdicts=report_got_verdicts,
        runUnderDoubleTl=report_run_under_double_tl,
        doubleTlVerdicts=report_double_tl_verdicts,
        sanitizerWarnings=report_sanitizer_warnings,
        verification=verification,
    )


def _print_solution_outcome(
    solution: Solution,
    skeleton: SolutionReportSkeleton,
    evals: List[Evaluation],
    console: rich.console.Console,
    verification: VerificationLevel = VerificationLevel.NONE,
    subset: bool = False,
    print_message: bool = True,
) -> bool:
    report = get_solution_outcome_report(
        solution, skeleton, evals, verification, subset
    )
    if not report.ok:
        issue_stack.add_issue(FailedSolutionIssue(solution))
    console.print(report.get_outcome_markup(print_message))
    return report.ok


def consume_and_key_evaluation_items(
    items: Iterable[EvaluationItem],
    skeleton: SolutionReportSkeleton,
) -> StructuredEvaluation:
    """
    Consumes EvaluationItems from a run_solutions call and build a view
    with them, possibly marking with optional unprocessed items.
    """
    res = skeleton.empty_structured_evaluation()

    for item in items:
        res[str(item.solution.path)][item.testcase_entry.group][
            item.testcase_entry.index
        ] = item.eval

    return res


def _print_solution_header(
    solution: SolutionSkeleton,
    console: rich.console.Console,
):
    console.print(f'{href(solution.path)}', end=' ')
    console.print(f'({href(solution.runs_dir, style="info")})')


@dataclasses.dataclass
class SolutionTiming:
    time: int
    solution: Solution


@dataclasses.dataclass
class TimingSummary:
    slowest_good: Optional[SolutionTiming] = None
    fastest_slow: Optional[SolutionTiming] = None

    def add_good(self, time: int, solution: Solution):
        if self.slowest_good is None or time > self.slowest_good.time:
            self.slowest_good = SolutionTiming(time, solution)

    def add_slow(self, time: int, solution: Solution):
        if self.fastest_slow is None or time < self.fastest_slow.time:
            self.fastest_slow = SolutionTiming(time, solution)

    def print(self, console: rich.console.Console, tl: Optional[int] = None):
        if self.slowest_good is not None:
            console.print(
                f'Slowest [success]OK[/success] solution: {self.slowest_good.time} ms, {href(self.slowest_good.solution.path)}'
            )
        if self.fastest_slow is not None:
            fastest_slow = self.fastest_slow.time
            if tl is not None and self.fastest_slow.time > tl:
                fastest_slow = f'>{tl}'
            console.print(
                f'Fastest [error]slow[/error] solution: {fastest_slow} ms, {href(self.fastest_slow.solution.path)}'
            )


async def _print_timing(
    console: rich.console.Console,
    skeleton: SolutionReportSkeleton,
    evaluations: StructuredEvaluation,
):
    summary = TimingSummary()
    summary_per_language = collections.defaultdict(TimingSummary)
    tls_per_language = {}
    all_tls = set()
    for solution in skeleton.solutions:
        all_evals: List[Evaluation] = []
        for evals in evaluations[str(solution.path)].values():
            all_evals.extend([await eval() for eval in evals if eval is not None])
        if not all_evals:
            continue

        # Get solution TL.
        solution_time = _get_evals_time_in_ms(all_evals)
        solution_tls = [
            eval.log.metadata.limits.get_expanded_tl()
            for eval in all_evals
            if eval.log.metadata is not None
        ]
        solution_tls = [tl for tl in solution_tls if tl is not None]

        solution_tl = 0
        if solution_tls:
            solution_tl = min(solution_tls)
        else:
            limits = skeleton.get_solution_limits(solution)
            if limits.time is None:
                limits = skeleton.get_solution_limits_from_disk(solution)
            assert limits.time is not None
            solution_tl = limits.time
            if limits.isDoubleTL:
                solution_tl = solution_tl * 2
        all_tls.add(solution_tl)
        for eval in all_evals:
            if eval.log.get_run_language() is not None:
                tls_per_language[eval.log.get_run_language()] = solution_tl

        # Get solution timings.
        if solution.outcome.match(Outcome.ACCEPTED):
            summary.add_good(solution_time, solution)
            summary_per_language[solution.language].add_good(solution_time, solution)
        if solution.outcome.is_slow():
            summary.add_slow(solution_time, solution)
            summary_per_language[solution.language].add_slow(solution_time, solution)

    if summary.slowest_good is None and summary.fastest_slow is None:
        return

    all_languages = set(summary_per_language)
    all_tl = min(all_tls) if all_tls else None
    console.print('[status]Timing summary:[/status]')

    if len(all_languages) <= 1 or len(all_tls) <= 1:
        summary.print(console, tl=all_tl)
        return

    # Otherwise, print per language.
    for lang in sorted(all_languages):
        console.print(f'[status]{lang}[/status]')
        summary_per_language[lang].print(
            console, tl=tls_per_language.get(lang) or all_tl
        )
        console.print()


def _length_markup(markup: str) -> int:
    text = rich.markup.render(markup)
    return text.cell_len


def _length_pointwise(ls: Iterable[str]) -> Tuple[int, ...]:
    return tuple(_length_markup(x) for x in ls)


def _max_pointwise(ls: Iterable[Tuple[int, ...]]) -> Tuple[int, ...]:
    return tuple(max(x) for x in zip(*ls))


def _get_indented_text(s: str, width: int):
    text = rich.markup.render(s)
    text.align('right', width=width)
    return text


def _render_padded_rows(
    rows: List[List[Tuple[str, ...]]],
) -> List[List[rich.text.Text]]:
    max_widths_per_column = [
        _max_pointwise(_length_pointwise(cell) for cell in col) for col in zip(*rows)
    ]
    res = []
    for row in rows:
        acc_row = []
        for i, cell in enumerate(row):
            acc_row.append(
                rich.text.Text(' ').join(
                    _get_indented_text(item, width)
                    for item, width in zip(cell, max_widths_per_column[i])
                )
            )
        res.append(acc_row)
    return res


async def _render_detailed_group_table(
    group: TestcaseGroup,
    skeleton: SolutionReportSkeleton,
    structured_evaluations: StructuredEvaluation,
    console: rich.console.Console,
    verification: VerificationLevel = VerificationLevel.NONE,
):
    group_skeleton = skeleton.find_group_skeleton(group.name)
    assert group_skeleton is not None

    async def generate_table(
        structured_evaluation: StructuredEvaluation, group_name: str
    ) -> rich.table.Table:
        table = rich.table.Table()
        for solution in skeleton.solutions:
            table.add_column(f'{href(solution.path)}', justify='full')

        padded_rows = []

        evals_per_solution = collections.defaultdict(list)
        for tc, _ in enumerate(group_skeleton.testcases):
            row = []
            for solution in skeleton.solutions:
                eval = structured_evaluation[str(solution.path)][group_name][tc]
                if eval is None:
                    row.append((f'[info]#{tc}[/info]', '', '...', '', '', ''))
                    continue
                eval = eval.peek()
                if eval is None:
                    row.append((f'[info]#{tc}[/info]', '', '...', '', '', ''))
                    continue

                evals_per_solution[str(solution.path)].append(eval)

                verdict = get_testcase_markup_verdict(eval)
                limits = skeleton.get_solution_limits(solution)
                time = get_capped_evals_formatted_time(limits, [eval], verification)
                memory = get_evals_formatted_memory([eval])
                full_item = (f'[info]#{tc}[/info]', verdict, time, '/', memory, '')
                if eval.result.sanitizer_warnings:
                    full_item = (*full_item[:-1], '[warning]*[/warning]')

                row.append(full_item)
            padded_rows.append(row)

        if padded_rows:
            summary_row = []
            for solution in skeleton.solutions:
                evals = evals_per_solution[str(solution.path)]
                non_null_evals = [eval for eval in evals if eval is not None]
                if not non_null_evals:
                    summary_row.append('...')
                    continue
                limits = skeleton.get_solution_limits(solution)
                formatted_time = get_capped_evals_formatted_time(
                    limits, non_null_evals, verification
                )
                formatted_memory = get_evals_formatted_memory(non_null_evals)
                worst_outcome = get_worst_outcome(non_null_evals)
                verdict = get_outcome_markup_verdict(worst_outcome)
                summary_row.append(
                    ('', verdict, formatted_time, '/', formatted_memory, '')
                )
            padded_rows.append(summary_row)

        for row in _render_padded_rows(padded_rows):
            table.add_row(*row)

        if padded_rows:
            table.rows[-2].end_section = True
        return table

    with rich.live.Live(
        await generate_table(skeleton.empty_structured_evaluation(), group.name),
        refresh_per_second=5,
        console=console,
    ) as live:
        for solution in skeleton.solutions:
            for tc, _ in enumerate(group_skeleton.testcases):
                eval = structured_evaluations[str(solution.path)][group.name][tc]
                if eval is None:
                    continue
                await eval()
                live.update(await generate_table(structured_evaluations, group.name))
                live.refresh()


async def _print_detailed_run_report(
    result: RunSolutionResult,
    console: rich.console.Console,
    structured_evaluations: StructuredEvaluation,
    timing: bool = True,
    verification: VerificationLevel = VerificationLevel.NONE,
):
    for group in result.skeleton.groups:
        console.print(f'[bold][status]{group.name}[/status][/bold]')

        await _render_detailed_group_table(
            package.get_testgroup(group.name),
            result.skeleton,
            structured_evaluations,
            console,
            verification=verification,
        )
        continue

    ok = True
    for solution in result.skeleton.solutions:
        all_evals = []
        for evals in structured_evaluations[str(solution.path)].values():
            all_evals.extend(evals)

        # Resolve futures.
        all_evals = [await eval() for eval in all_evals if eval is not None]
        _print_solution_header(solution, console)
        cur_ok = _print_solution_outcome(
            solution,
            result.skeleton,
            all_evals,
            console,
            verification=verification,
        )
        ok = ok and cur_ok
        console.print()

    console.print()

    if timing:
        await _print_timing(
            console,
            result.skeleton,
            structured_evaluations,
        )
    return ok


def _print_limits(limits: Dict[str, Limits]):
    console.console.print(
        '[bold][success]Running with the following limits (per language):[/success][/bold]'
    )
    for lang, limit in limits.items():
        extracted_from = ' (extracted from package)'
        if limit.profile:
            extracted_from = f' (extracted from profile [item]{limit.profile}[/item])'
        console.console.print(f'[bold][status]{lang}[/status][/bold]{extracted_from}')
        time = (
            '<No time limit>' if limit.time is None else get_formatted_time(limit.time)
        )
        memory = (
            '<No memory limit>'
            if limit.memory is None
            else get_formatted_memory(limit.memory * 1024 * 1024)
        )
        console.console.print(f'Time: {time}')
        console.console.print(f'Memory: {memory}')
        if limit.isDoubleTL:
            console.console.print('[warning]Running with 2*TL[/warning]')
    console.console.print()


async def print_run_report(
    result: RunSolutionResult,
    console: rich.console.Console,
    verification: VerificationLevel,
    detailed: bool = False,
    timing: bool = True,
    skip_printing_limits: bool = False,
) -> bool:
    if not skip_printing_limits:
        _print_limits(result.skeleton.limits)

    structured_evaluations = consume_and_key_evaluation_items(
        result.items, result.skeleton
    )
    if detailed:
        return await _print_detailed_run_report(
            result,
            console,
            structured_evaluations,
            verification=verification,
            timing=timing,
        )

    ok = True
    single_solution = len(result.skeleton.solutions) == 1

    for solution in result.skeleton.solutions:
        _print_solution_header(solution, console)
        if single_solution:
            console.print()
        solution_evals = []
        limits = result.skeleton.get_solution_limits(solution)
        for group in result.skeleton.groups:
            if not single_solution:
                console.print(f'[bold][status]{group.name}[/status][/bold] ', end='')
            group_evals = []
            for i, _ in enumerate(group.testcases):
                eval = structured_evaluations[str(solution.path)][group.name][i]
                if eval is None:
                    continue
                eval = await eval()
                if single_solution:
                    console.print(get_testcase_markup_verdict(eval), end=' ')
                    console.print(f'{group.name}/{i}', end='')
                    if eval.result.sanitizer_warnings:
                        console.print('[warning]*[/warning]', end='')
                    time = get_capped_evals_formatted_time(limits, [eval], verification)
                    memory = get_evals_formatted_memory([eval])
                    console.print(f' ({time}, {memory})', end='')
                    checker_msg = eval.result.message
                    if checker_msg:
                        checker_msg = get_truncated_message(checker_msg, 150)
                        console.print(
                            f': [i]{utils.escape_markup(checker_msg)}[/i]', end=''
                        )
                else:
                    console.print(f'{i}/', end='')
                    console.print(get_testcase_markup_verdict(eval), end='')
                    if eval.result.sanitizer_warnings:
                        console.print('[warning]*[/warning]', end='')

                console.print('', end='\n' if single_solution else ' ')
                group_evals.append(eval)
                solution_evals.append(eval)

            if single_solution:
                console.print(f'  [status]{group.name}[/status]', end=' ')
            console.print(
                f'({get_capped_evals_formatted_time(limits, group_evals, verification)}, {get_evals_formatted_memory(group_evals)})',
                end='',
            )
            console.print()
            if single_solution:
                console.print()

        cur_ok = _print_solution_outcome(
            solution,
            result.skeleton,
            solution_evals,
            console,
            verification=verification,
            print_message=not single_solution,
        )
        ok = ok and cur_ok
        console.print()

    if not single_solution:
        await _print_timing(
            console,
            result.skeleton,
            structured_evaluations,
        )

    return ok



import pathlib

import mkdocs_gen_files

from rbx.box.contest.schema import Contest
from rbx.box.environment import Environment
from rbx.box.package import Package
from rbx.box.presets.lock_schema import PresetLock
from rbx.box.presets.schema import Preset
from rbx.box.schema import LimitsProfile
from rbx.box.statements.schema import Statement
from rbx.utils import dump_schema_str

models = [Package, Environment, Contest, Preset, PresetLock, Statement, LimitsProfile]

for model in models:
    path = pathlib.Path('schemas') / f'{model.__name__}.json'
    with mkdocs_gen_files.open(str(path), 'w') as f:
        f.write(dump_schema_str(model))



import pathlib
from typing import Any, Dict, List, Optional, Tuple

from rbx.grading import grading_context, profiling, steps
from rbx.grading.caching import DependencyCache, NoCacheException
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    GradingArtifacts,
    GradingLogsHolder,
    RunLog,
    RunLogMetadata,
)


def _get_prefixed_cacheable_params(
    params: Dict[str, Any], prefix: str
) -> Dict[str, Any]:
    return {f'{prefix}.{k}': v for k, v in params.items()}


def compile(
    commands: List[str],
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
):
    artifacts.logs = GradingLogsHolder()

    ok = True
    cached_profile = profiling.Profiler('steps.compile[cached]', start=True)
    with dependency_cache(
        commands, [artifacts], params.get_cacheable_params()
    ) as is_cached:
        if not is_cached:
            with profiling.Profiler('steps.compile'):
                profiling.add_to_counter('steps.compile')
                ok = steps.compile(
                    commands=commands,
                    params=params,
                    artifacts=artifacts,
                    sandbox=sandbox,
                )
                if not ok:
                    raise NoCacheException()
        else:
            cached_profile.stop()
            profiling.add_to_counter('steps.compile[cached]')

    return ok


async def run(
    command: str,
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
    metadata: Optional[RunLogMetadata] = None,
) -> Optional[RunLog]:
    artifacts.logs = GradingLogsHolder()

    cacheable_params = params.get_cacheable_params()
    if metadata is not None and metadata.retryIndex is not None:
        cacheable_params['__retry_index__'] = metadata.retryIndex

    with grading_context.cache_level(
        grading_context.CacheLevel.NO_CACHE,
        when=grading_context.is_compilation_only,
    ):
        cached_profile = profiling.Profiler('steps.run[cached]', start=True)
        with dependency_cache([command], [artifacts], cacheable_params) as is_cached:
            if not is_cached:
                with profiling.Profiler('steps.run'):
                    profiling.add_to_counter('steps.run')
                    await steps.run(
                        command=command,
                        params=params,
                        artifacts=artifacts,
                        sandbox=sandbox,
                        metadata=metadata,
                    )
            else:
                cached_profile.stop()
                profiling.add_to_counter('steps.run[cached]')
    return artifacts.logs.run


async def run_coordinated(
    interactor: steps.CoordinatedRunParams,
    solution: steps.CoordinatedRunParams,
    artifacts: GradingArtifacts,
    sandbox: SandboxBase,
    dependency_cache: DependencyCache,
    merged_capture: Optional[pathlib.Path] = None,
) -> Tuple[Optional[RunLog], Optional[RunLog]]:
    artifacts.logs = GradingLogsHolder()

    cacheable_params = {
        **_get_prefixed_cacheable_params(
            interactor.params.get_cacheable_params(), 'interactor'
        ),
        **_get_prefixed_cacheable_params(
            solution.params.get_cacheable_params(), 'solution'
        ),
    }

    if interactor.metadata is not None and interactor.metadata.retryIndex is not None:
        cacheable_params['interactor.__retry_index__'] = interactor.metadata.retryIndex
    if solution.metadata is not None and solution.metadata.retryIndex is not None:
        cacheable_params['solution.__retry_index__'] = solution.metadata.retryIndex

    with grading_context.cache_level(
        grading_context.CacheLevel.NO_CACHE,
        when=grading_context.is_compilation_only,
    ):
        cached_profile = profiling.Profiler('steps.run_coordinated[cached]', start=True)
        with dependency_cache(
            [interactor.command, solution.command],
            [artifacts],
            cacheable_params,
        ) as is_cached:
            if not is_cached:
                with profiling.Profiler('steps.run_coordinated'):
                    profiling.add_to_counter('steps.run_coordinated')
                    await steps.run_coordinated(
                        interactor,
                        solution,
                        artifacts,
                        sandbox,
                        merged_capture=merged_capture,
                    )
            else:
                cached_profile.stop()
                profiling.add_to_counter('steps.run_coordinated[cached]')

    return (
        artifacts.logs.interactor_run,
        artifacts.logs.run,
    )



import importlib.resources
import pathlib
import re
from typing import Any, Dict, List, Optional

import typer
import typer.core
from typing_extensions import Annotated

from rbx import config
from rbx.config import get_config


def _get_language_options():
    return sorted(get_config().languages.keys())


def _get_language_default():
    return get_config().defaultLanguage


def _get_problem_options():
    options = set()
    # all_problems = metadata.find_problems()
    # for problem in all_problems:
    #     options.add(problem.code)
    #     options.update(problem.aliases)
    return sorted(options)


def _list_files(path: pathlib.Path) -> List[str]:
    if not path.is_dir():
        return []
    return [file.name for file in path.iterdir() if file.is_file()]


def _get_checker_options():
    options = set()
    with importlib.resources.as_file(
        importlib.resources.files('rbx') / 'resources' / 'checkers'
    ) as file:
        options.update(_list_files(file))

    options.update(_list_files(config.get_app_path() / 'checkers'))
    options.remove('boilerplate.cpp')
    return sorted(options)


Timelimit = Annotated[
    int,
    typer.Option(
        '--timelimit',
        '-t',
        help='Time limit in milliseconds.',
        prompt='Time limit (ms)',
    ),
]
Memorylimit = Annotated[
    int,
    typer.Option(
        '--memorylimit',
        '-m',
        help='Memory limit in megabytes.',
        prompt='Memory limit (MB)',
    ),
]
Multitest = Annotated[
    Optional[bool],
    typer.Option(
        '--multitest',
        '-m',
        is_flag=True,
        help='Whether this problem have multiple tests per file.',
        prompt='Multitest?',
    ),
]
Language = Annotated[
    str,
    typer.Option(
        '--language',
        '--lang',
        '-l',
        help='Language to use.',
        prompt='Language',
        default_factory=_get_language_default,
        autocompletion=_get_language_options,
    ),
]
LanguageWithDefault = Annotated[
    Optional[str],
    typer.Option(
        '--language',
        '--lang',
        '-l',
        help='Language to use.',
        autocompletion=_get_language_options,
    ),
]
Problem = Annotated[str, typer.Argument(autocompletion=_get_problem_options)]

ProblemOption = Annotated[
    Optional[str], typer.Option('--problem', '-p', autocompletion=_get_problem_options)
]

TestcaseIndex = Annotated[Optional[int], typer.Option('--index', '--idx', '-i')]

Checker = Annotated[
    str,
    typer.Argument(
        autocompletion=_get_checker_options, help='Path to a testlib checker file.'
    ),
]


def parse_dictionary(value: Optional[str]) -> Dict[str, Any]:
    if value is None:
        return {}
    res = {}
    for item in value.split(','):
        key, value = item.split('=', 1)
        res[key] = value
    return res


def parse_dictionary_items(items: Optional[List[str]]) -> Dict[str, Any]:
    if items is None:
        return {}
    res = {}
    for item in items:
        key, value = item.split('=', 1)
        res[key] = value
    return res


class AliasGroup(typer.core.TyperGroup):
    _CMD_SPLIT_P = re.compile(r', ?')

    def get_command(self, ctx, cmd_name):
        cmd_name = self._group_cmd_name(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _group_cmd_name(self, default_name):
        for cmd in self.commands.values():
            if cmd.name and default_name in self._CMD_SPLIT_P.split(cmd.name):
                return cmd.name
        return default_name



import contextlib
import enum
import functools
import json
import os
import os.path
import pathlib
import re
import shutil
import subprocess
import sys
from typing import Any, Optional, Type, TypeVar, Union

import rich
import rich.markup
import rich.prompt
import rich.status
import ruyaml
import semver
import typer
import yaml
from pydantic import BaseModel
from rich import text
from rich.highlighter import JSONHighlighter

from rbx import __version__
from rbx.console import console

T = TypeVar('T', bound=BaseModel)
APP_NAME = 'rbx'
PIP_NAME = 'rbx.cp'
PathOrStr = Union[pathlib.Path, str]


class SemVerCompatibility(enum.Enum):
    COMPATIBLE = 'compatible'
    OUTDATED = 'outdated'
    BREAKING_CHANGE = 'breaking_change'


def get_version() -> str:
    return __version__.__version__


def get_semver() -> semver.VersionInfo:
    return semver.VersionInfo.parse(get_version())


def get_upgrade_command(
    version: Optional[Union[str, semver.VersionInfo]] = None,
) -> str:
    parsed_version = (
        semver.VersionInfo.parse(version) if isinstance(version, str) else version
    ) or get_semver()
    return f'pipx install --upgrade {PIP_NAME}@{parsed_version.major}'


def check_version_compatibility_between(
    installed: str,
    required: str,
) -> SemVerCompatibility:
    installed_version = semver.VersionInfo.parse(installed)
    required_version = semver.VersionInfo.parse(required)
    if installed_version < required_version:
        return SemVerCompatibility.OUTDATED
    if installed_version.major > required_version.major:
        return SemVerCompatibility.BREAKING_CHANGE
    return SemVerCompatibility.COMPATIBLE


def check_version_compatibility(required: str) -> SemVerCompatibility:
    installed = get_version()
    return check_version_compatibility_between(installed, required)


def print_open_fd_count(id: Optional[str] = None) -> int:
    import psutil

    try:
        open_fds = get_open_fds()
        print(f'Number of opened file descriptors for {id or "..."}: {open_fds}')
    except psutil.AccessDenied:
        print('Access denied. Run with appropriate permissions (e.g., sudo) if needed.')
    except Exception as e:
        print(f'An error occurred: {e}')


class FdLeakDetector:
    def __init__(self, id: Optional[str] = None, diff: bool = False):
        self.open_fds = 0
        self.id = id
        self.diff = diff

    def __enter__(self):
        self.open_fds = get_open_fds()
        return self

    def __exit__(self, *args, **kwargs):
        open_fds = get_open_fds()
        if open_fds > self.open_fds:
            print(
                f'File descriptor leak detected for {self.id or "..."}: {open_fds - self.open_fds} new file descriptors opened'
            )
        elif self.diff and open_fds < self.open_fds:
            print(
                f'File descriptor diff detected for {self.id or "..."}: {self.open_fds - open_fds} file descriptors closed'
            )
        self.open_fds = open_fds


def create_and_write(path: pathlib.Path, *args, **kwargs):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(*args, **kwargs)


def highlight_str(s: str) -> text.Text:
    txt = text.Text(s)
    JSONHighlighter().highlight(txt)
    return txt


def escape_markup(s: str) -> str:
    return rich.markup.escape(s, _escape=re.compile(r'(\\*)(\[)').sub)


def abspath(path: pathlib.Path) -> pathlib.Path:
    return pathlib.Path(os.path.abspath(path))


def relpath(path: pathlib.Path, base: pathlib.Path) -> pathlib.Path:
    if sys.version_info >= (3, 12):
        return path.relative_to(base, walk_up=True)
    else:
        return pathlib.Path(os.path.relpath(path, base))


def highlight_json_obj(obj: Any) -> text.Text:
    js = json.dumps(obj)
    return highlight_str(js)


def normalize_with_underscores(s: str) -> str:
    res = s.replace(' ', '_').replace('.', '_').strip('_')
    final = []

    last = ''
    for c in res:
        if c == '_' and last == c:
            continue
        last = c
        final.append(c)
    return ''.join(final)


def get_app_path() -> pathlib.Path:
    app_dir = typer.get_app_dir(APP_NAME)
    return pathlib.Path(app_dir)


def dump_schema_str(model: Type[BaseModel]) -> str:
    return json.dumps(model.model_json_schema(), indent=4)


def dump_schema(model: Type[BaseModel], path: pathlib.Path):
    path.parent.mkdir(parents=True, exist_ok=True)
    schema = dump_schema_str(model)
    path.write_text(schema)


def ensure_schema(model: Type[BaseModel]) -> pathlib.Path:
    path = get_app_path() / 'schemas' / f'{model.__name__}.json'
    dump_schema(model, path)
    return abspath(path)


def model_json(model: BaseModel) -> str:
    return model.model_dump_json(indent=4, exclude_unset=True, exclude_none=True)


def uploaded_schema_path(model: Type[BaseModel]) -> str:
    return f'https://rsalesc.github.io/rbx/schemas/{model.__name__}.json'


def model_to_yaml(model: BaseModel, **kwargs) -> str:
    """Convert model to YAML string with proper boolean handling.

    This function works around Pydantic's issue where Union[str, int, float, bool]
    fields convert booleans to floats when using mode='json'.
    """
    # Use regular dump to preserve boolean types
    data = model.model_dump(exclude_unset=True, exclude_none=True)

    # Ensure the result is JSON-serializable by converting any non-JSON types
    json_safe_data = _ensure_json_serializable(data)

    # Add schema path comment and convert to YAML
    path = uploaded_schema_path(model.__class__)
    schema_comment = f'# yaml-language-server: $schema={path}\n\n'

    yaml_content = yaml.safe_dump(
        json_safe_data, sort_keys=False, allow_unicode=True, **kwargs
    )

    return schema_comment + yaml_content


def _ensure_json_serializable(obj):
    """Recursively ensure an object is JSON-serializable while preserving booleans."""
    from datetime import date, datetime
    from enum import Enum
    from pathlib import Path
    from uuid import UUID

    from rbx.autoenum import AutoEnum

    if isinstance(obj, dict):
        return {k: _ensure_json_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_ensure_json_serializable(item) for item in obj]
    elif isinstance(obj, tuple):
        return [_ensure_json_serializable(item) for item in obj]
    elif isinstance(obj, set):
        return [_ensure_json_serializable(item) for item in obj]
    elif isinstance(obj, AutoEnum):
        return str(obj)
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    elif isinstance(obj, Path):
        return str(obj)
    else:
        # For any other type, try to convert to string
        return str(obj)


def model_from_yaml(model: Type[T], s: str) -> T:
    return model(**yaml.safe_load(s))


def validate_field(model: Type[T], field: str, value: Any):
    model.__pydantic_validator__.validate_assignment(
        model.model_construct(), field, value
    )


def save_ruyaml(path: pathlib.Path, yml: ruyaml.YAML, data: ruyaml.Any):
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w') as f:
        yml.dump(data, f)


@functools.cache
def get_empty_sentinel_path() -> pathlib.Path:
    path = get_app_path() / '.empty'
    path.write_text('')
    return path


@contextlib.contextmanager
def no_progress(status: Optional[rich.status.Status]):
    if status:
        status.stop()
    yield
    if status:
        status.start()


def confirm_on_status(status: Optional[rich.status.Status], *args, **kwargs) -> bool:
    with no_progress(status):
        res = rich.prompt.Confirm.ask(*args, **kwargs, console=console)
    return res


def get_open_fds() -> int:
    import psutil

    try:
        current_process = psutil.Process()
        open_fds = current_process.num_fds()
        return open_fds
    except psutil.AccessDenied:
        return 0
    except Exception as e:
        print(f'An error occurred: {e}')
        return 0

    # fds = []
    # soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    # for fd in range(0, soft):
    #     try:
    #         fcntl.fcntl(fd, fcntl.F_GETFD)
    #     except IOError:
    #         continue
    #     fds.append(fd)
    # return fds


def command_exists(command):
    try:
        subprocess.run(
            [command], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return True
    except FileNotFoundError:
        return False
    except subprocess.CalledProcessError:
        return True


@contextlib.contextmanager
def new_cd(x: pathlib.Path):
    d = os.getcwd()

    # This could raise an exception, but it's probably
    # best to let it propagate and let the caller
    # deal with it, since they requested x
    os.chdir(x)

    try:
        yield

    finally:
        # This could also raise an exception, but you *really*
        # aren't equipped to figure out what went wrong if the
        # old working directory can't be restored.
        os.chdir(d)


def _safe_match(matcher, path):
    try:
        return matcher(path)
    except ValueError:
        return False


def copytree_honoring_gitignore(
    src: pathlib.Path, dst: pathlib.Path, extra_gitignore: Optional[str] = None
):
    from gitignore_parser import parse_gitignore, parse_gitignore_str

    ignore_matchers = []

    if extra_gitignore is not None:
        ignore_matchers.append(parse_gitignore_str(extra_gitignore, base_dir=src))

    for file in src.rglob('.gitignore'):
        if file.is_file():
            ignore_matchers.append(parse_gitignore(file))

    # TODO: use recursive walk
    for file in src.rglob('*'):
        matching_file = file
        ignored = False
        while matching_file.is_relative_to(src):
            if any(
                _safe_match(ignore_matcher, matching_file)
                for ignore_matcher in ignore_matchers
            ):
                ignored = True
                break
            matching_file = matching_file.parent
        if ignored:
            continue
        rel = relpath(file, src)
        if file.is_file():
            write_to = dst / rel
            write_to.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(file, write_to)


class StatusProgress(rich.status.Status):
    _message: str
    processed: int
    keep: bool

    def __init__(
        self, message: str, formatted_message: Optional[str] = None, keep: bool = False
    ):
        self._message = formatted_message or message
        self.keep = keep
        self.processed = 0
        super().__init__(message.format(processed=0), console=console)
        self.start()

    def __enter__(self):
        super().__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)
        if self.keep:
            console.print(self._message.format(processed=self.processed))

    def update_with_progress(self, processed: int):
        self.processed = processed
        self.update(self._message.format(processed=processed))

    def step(self, delta: int = 1):
        self.processed += delta
        self.update_with_progress(self.processed)



import typing

from rbx.box.environment import get_environment, get_language
from rbx.box.packaging.boca.extension import BocaLanguage, BocaLanguageExtension


def get_rbx_language_from_boca_language(boca_language: BocaLanguage) -> str:
    for language in get_environment().languages:
        language_extension = language.get_extension_or_default(
            language.name, BocaLanguageExtension
        )
        if language_extension.bocaLanguage == boca_language:
            return language.name
    return boca_language


def get_boca_language_from_rbx_language(rbx_language: str) -> BocaLanguage:
    language = get_language(rbx_language)
    language_extension = language.get_extension_or_default(
        'boca', BocaLanguageExtension
    )
    if language_extension.bocaLanguage:
        return typing.cast(BocaLanguage, language_extension.bocaLanguage)
    if rbx_language.lower() in typing.get_args(BocaLanguage):
        return typing.cast(BocaLanguage, rbx_language.lower())
    raise ValueError(f'No Boca language found for Rbx language {rbx_language}')



import pathlib
import shutil
from typing import List, Optional

import typer

from rbx import console, utils
from rbx.box import lang
from rbx.box.packaging.importer import BaseImporter
from rbx.box.packaging.polygon.xml_schema import File, Problem, Statement, Testset
from rbx.box.schema import CodeItem, Interactor, Package, TaskType, TestcaseGroup
from rbx.box.statements.schema import Statement as BoxStatement
from rbx.box.statements.schema import StatementType


def _get_main_testset(problem: Problem) -> Testset:
    for testset in problem.judging.testsets:
        if testset.name == 'tests':
            return testset
    console.console.print(
        '[error][item]tests[/item] testset not found[/error]',
    )
    raise typer.Exit(1)


def _get_pdf_statements(problem: Problem) -> List[Statement]:
    statements = []
    for statement in problem.statements:
        if statement.type == 'application/pdf':
            statements.append(statement)
    return statements


def _get_statement_path(statement: Statement) -> pathlib.Path:
    return pathlib.Path('statements') / f'{statement.language}.pdf'


def _populate_tests(
    testset: Testset, pkg: Package, pkg_path: pathlib.Path, into_path: pathlib.Path
):
    if not testset.answerPattern:
        console.console.print(
            '[error][item]answer pattern[/item] not found for testset[/error]',
        )
        raise typer.Exit(1)

    for d, test in enumerate(testset.tests):
        folder_name = 'tests/samples' if test.sample else 'tests/tests'
        i = d + 1

        input_path = pathlib.Path(testset.inputPattern % i)
        dest_input_path = into_path / folder_name / f'{i:03d}.in'
        dest_input_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(pkg_path / input_path, dest_input_path)

        answer_path = pathlib.Path(testset.answerPattern % i)
        dest_answer_path = into_path / folder_name / f'{i:03d}.ans'
        dest_answer_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(pkg_path / answer_path, dest_answer_path)

    pkg.testcases = [
        TestcaseGroup(
            name='samples',
            testcaseGlob='tests/samples/*.in',
        ),
        TestcaseGroup(
            name='tests',
            testcaseGlob='tests/tests/*.in',
        ),
    ]


def _populate_titles(problem: Problem, pkg: Package):
    titles = {}
    for name in problem.names:
        iso639_code = lang.lang_to_code(name.language)
        titles[iso639_code] = name.value
    pkg.titles = titles


def _populate_statements(
    problem: Problem,
    pkg: Package,
    pkg_path: pathlib.Path,
    into_path: pathlib.Path,
    main_language: Optional[str] = None,
):
    name_per_language = {name.language: name for name in problem.names}
    pdf_statements = _get_pdf_statements(problem)
    pkg_statements = []
    found_main = False

    for statement in pdf_statements:
        statement_path = into_path / _get_statement_path(statement)
        statement_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(pkg_path / statement.path, statement_path)

        iso639_code = lang.lang_to_code(statement.language)

        pkg_statement = BoxStatement(
            name=f'statement-{statement.language}',
            language=iso639_code,
            path=_get_statement_path(statement),
            type=StatementType.PDF,
        )

        if (
            main_language is not None
            and main_language == iso639_code
            and not found_main
        ):
            # If main statement, add it to the front of the list
            pkg_statements = [pkg_statement] + pkg_statements
            found_main = True
            continue

        name = name_per_language.get(statement.language)
        if name is not None and name.main and not found_main:
            # If main statement, add it to the front of the list
            pkg_statements = [pkg_statement] + pkg_statements
            found_main = True
            continue

        pkg_statements.append(pkg_statement)

    pkg.statements = pkg_statements

    if main_language is not None and not found_main:
        console.console.print(
            f'[error]Main statement of language [item]{main_language}[/item] not found.[/error]',
        )
        console.console.print(
            'If you want no statement in your imported package, '
            'leave the [item]--main-language[/item] flag unset.'
        )
        raise typer.Exit(1)

    if not pkg_statements:
        console.console.print(
            '[warning]Imported problem has no statements. Continuing without a statement.[/warning]',
        )


def _is_cpp_source(source: File) -> bool:
    if source.type is None:
        return False
    return 'cpp' in source.type


def _copy_checker(
    problem: Problem, pkg: Package, pkg_path: pathlib.Path, into_path: pathlib.Path
):
    if problem.checker is None:
        return
    if problem.checker.type != 'testlib' or not _is_cpp_source(problem.checker.source):
        console.console.print(
            f'[error][item]checker type[/item] not supported: [item]{problem.checker.type}[/item][/error]',
        )
        raise typer.Exit(1)
    shutil.copy(pkg_path / problem.checker.source.path, into_path / 'checker.cpp')

    pkg.checker = CodeItem(
        path=pathlib.Path('checker.cpp'),
    )


def _copy_interactor(
    problem: Problem, pkg: Package, pkg_path: pathlib.Path, into_path: pathlib.Path
):
    if problem.interactor is None:
        return
    shutil.copy(pkg_path / problem.interactor.source.path, into_path / 'interactor.cpp')

    if not _is_cpp_source(problem.interactor.source):
        console.console.print(
            f'[error]Only C++ interactor is supported, got [item]{problem.interactor.source.type}[/item][/error]',
        )
        raise typer.Exit(1)

    pkg.type = TaskType.COMMUNICATION
    pkg.interactor = Interactor(
        path=pathlib.Path('interactor.cpp'),
        legacy=True,
    )


def _copy_headers(
    problem: Problem, pkg: Package, pkg_path: pathlib.Path, into_path: pathlib.Path
):
    headers = []
    for file in problem.files:
        if file.type is None or not file.type.startswith('h.'):
            continue
        header_path = pkg_path / file.path
        dest_path = into_path / header_path.name
        if header_path.name == 'rbx.h':
            dest_path = into_path / 'rbx.override.h'
        shutil.copy(header_path, dest_path)
        headers.append(dest_path.name)

    if pkg.checker is not None:
        pkg.checker.compilationFiles = headers

    if pkg.interactor is not None:
        pkg.interactor.compilationFiles = headers


class PolygonImporter(BaseImporter):
    def __init__(self, main_language: Optional[str]):
        self.main_language = main_language

    @classmethod
    def name(cls) -> str:
        return 'polygon'

    async def import_package(self, pkg_path: pathlib.Path, into_path: pathlib.Path):
        problem_xml = pkg_path / 'problem.xml'
        if not problem_xml.exists():
            console.console.print(
                '[error][item]problem.xml[/item] not found[/error]',
            )
            raise typer.Exit(1)

        problem = Problem.from_xml(problem_xml.read_bytes())
        testset = _get_main_testset(problem)

        if testset.timelimit is None:
            testset.timelimit = 1000

        if testset.memorylimit is None:
            testset.memorylimit = 256 * 1024 * 1024

        pkg = Package(
            name=problem.short_name,
            timeLimit=testset.timelimit,
            memoryLimit=testset.memorylimit // (1024 * 1024),
            outputLimit=64 * 1024,
        )

        _populate_tests(testset, pkg, pkg_path, into_path)
        _populate_titles(problem, pkg)
        _populate_statements(problem, pkg, pkg_path, into_path, self.main_language)
        _copy_checker(problem, pkg, pkg_path, into_path)
        _copy_interactor(problem, pkg, pkg_path, into_path)
        _copy_headers(problem, pkg, pkg_path, into_path)

        (into_path / 'problem.rbx.yml').write_text(utils.model_to_yaml(pkg))



import asyncio
from typing import List

from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Footer, Header

from rbx.box.ui.captured_log import LogDisplay


class CommandScreen(Screen):
    BINDINGS = [('q', 'app.pop_screen', 'Back')]

    def __init__(self, command: List[str]):
        super().__init__()
        self.command = command

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield LogDisplay()

    async def _run_command(self):
        exitcode = await self.query_one(LogDisplay).capture(self.command)
        if exitcode != 0:
            self.query_one(LogDisplay).border_subtitle = f'Exit code: {exitcode}'
            return

        self.query_one(LogDisplay).border_subtitle = 'Finished'

    async def on_mount(self):
        self.query_one(LogDisplay).border_title = 'Command output'

        # Fire and forget.
        asyncio.create_task(self._run_command())



from typing import TYPE_CHECKING, Dict, TypeVar, Union

from deepmerge import always_merger
from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType


def NameField(**kwargs):
    return Field(
        pattern=r'^[a-zA-Z0-9][a-zA-Z0-9\-_]*$', min_length=3, max_length=32, **kwargs
    )


def FNameField(**kwargs):
    return Field(
        pattern=r'^[a-zA-Z0-9][a-zA-Z0-9\-_]*$', min_length=3, max_length=128, **kwargs
    )


T = TypeVar('T', bound=BaseModel)


def merge_pydantic_models(base: T, nxt: T) -> T:
    """Merge two Pydantic model instances.

    The attributes of 'base' and 'nxt' that weren't explicitly set are dumped into dicts
    using '.model_dump(exclude_unset=True)', which are then merged using 'deepmerge',
    and the merged result is turned into a model instance using '.model_validate'.

    For attributes set on both 'base' and 'nxt', the value from 'nxt' will be used in
    the output result.
    """
    base_dict = base.model_dump(exclude_unset=True)
    nxt_dict = nxt.model_dump(exclude_unset=True)
    merged_dict = always_merger.merge(base_dict, nxt_dict)
    return base.model_validate(merged_dict)


Primitive = Union[int, float, bool, str]
Vars = Dict[str, Primitive]
if TYPE_CHECKING:
    RecVars = Dict[str, Union[Primitive, 'RecVars']]
else:
    RecVars = TypeAliasType('RecVars', "Dict[str, Union[Primitive, 'RecVars']]")


def expand_var(value: Primitive) -> Primitive:
    if not isinstance(value, str):
        return value
    if value.startswith('\\'):
        return value[1:]
    if not value.startswith('py`') or not value.endswith('`'):
        return value
    res = eval(value[3:-1])
    for supported_type in [str, int, float, bool]:
        if isinstance(res, supported_type):
            return res

    raise TypeError(
        f'Variable with backticks should evaluate to a primitive Python type: {value}'
    )


def expand_vars(recvars: RecVars) -> Vars:
    vars = {}

    def solve(rec: RecVars, prefix: str) -> None:
        nonlocal vars
        for k, v in rec.items():
            if isinstance(v, dict):
                solve(v, f'{prefix}.{k}')
            else:
                key = f'{prefix}.{k}'.strip('.')
                vars[key] = expand_var(v)

    solve(recvars, '')
    return vars



import pathlib
import subprocess
from typing import List, Optional

import git
import semver


def get_repo_or_nil(
    root: pathlib.Path = pathlib.Path(), search_parent_directories: bool = False
) -> Optional[git.Repo]:
    try:
        return git.Repo(root, search_parent_directories=search_parent_directories)
    except git.InvalidGitRepositoryError:
        return None


def is_repo(path: pathlib.Path) -> bool:
    return get_repo_or_nil(path, search_parent_directories=False) is not None


def is_within_repo(path: pathlib.Path) -> bool:
    return get_repo_or_nil(path, search_parent_directories=True) is not None


def get_any_remote(repo: git.Repo) -> Optional[git.Remote]:
    for remote in repo.remotes:
        if remote.exists():
            return remote
    return None


def _parse_tag_from_ref(ref: str) -> str:
    return ref.split('/')[-1].split('^{}')[0]


def ls_remote_tags(uri: str) -> List[str]:
    completed_process = subprocess.run(
        ['git', 'ls-remote', '--tags', uri],
        check=True,
        capture_output=True,
        text=True,
    )
    return [
        _parse_tag_from_ref(line.split('\t')[1])
        for line in completed_process.stdout.split('\n')
        if line
    ]


def ls_version_remote_tags(uri: str) -> List[str]:
    tags = ls_remote_tags(uri)
    valid_tags = [tag for tag in tags if semver.Version.is_valid(tag)]
    return valid_tags


def latest_remote_tag(
    uri: str, before: Optional[str] = None, after: Optional[str] = None
) -> str:
    tags = ls_version_remote_tags(uri)
    if not tags:
        raise ValueError(f'No valid tags found for {uri}')
    if before is not None:
        tags = [
            tag
            for tag in tags
            if semver.VersionInfo.parse(tag) <= semver.VersionInfo.parse(before)
        ]
    if after is not None:
        tags = [
            tag
            for tag in tags
            if semver.VersionInfo.parse(tag) >= semver.VersionInfo.parse(after)
        ]
    return sorted(tags, key=semver.VersionInfo.parse)[-1]


def has_remote_tag(uri: str, tag: str) -> bool:
    tags = ls_remote_tags(uri)
    return tag in tags



#!/usr/bin/env python3
"""
Forked from https://github.com/RagnarGrootKoerkamp/BAPCtools/blob/master/bin/interactive.py

Takes a character and a file name as arguments.

Reads from stdin, and writes to stdout and stderr, with the given character prepended to
every line read from stdin.
"""

import sys

c = sys.argv[1]
extra = sys.argv[2]

new = True

with open(extra, 'w') as f:
    while True:
        rd = sys.stdin.read(1)
        if rd == '':
            break
        sys.stdout.write(rd)
        sys.stdout.flush()
        if new:
            sys.stderr.write(c)
        sys.stderr.write(rd)
        sys.stderr.flush()

        f.write(rd)
        new = rd == '\n'



import contextvars
import pathlib
from typing import Optional

import typer

from rbx import console, utils
from rbx.box import package
from rbx.box.environment import VerificationLevel
from rbx.box.schema import LimitsProfile
from rbx.grading.limits import Limits

profile_var = contextvars.ContextVar[Optional[str]]('profile', default=None)


def get_active_profile() -> Optional[str]:
    return profile_var.get()


class use_profile:
    def __init__(self, profile: Optional[str]):
        self.profile = profile
        self.token = None

    def __enter__(self):
        self.token = profile_var.set(self.profile)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.token is not None:
            profile_var.reset(self.token)


def _expand_limits_profile(
    limits_profile: LimitsProfile, root: pathlib.Path
) -> LimitsProfile:
    pkg = package.find_problem_package_or_die(root=root)
    res = LimitsProfile(
        timeLimit=pkg.timeLimit,
        memoryLimit=pkg.memoryLimit,
        outputLimit=pkg.outputLimit,
    )
    for language, modifier in pkg.modifiers.items():
        res.modifiers[language] = modifier.model_copy(deep=True)

    if limits_profile.inheritFromPackage:
        return res

    time_is_overridden = limits_profile.timeLimit is not None
    memory_is_overridden = limits_profile.memoryLimit is not None
    output_is_overridden = limits_profile.outputLimit is not None
    if time_is_overridden:
        res.timeLimit = limits_profile.timeLimit
    if memory_is_overridden:
        res.memoryLimit = limits_profile.memoryLimit
    if output_is_overridden:
        res.outputLimit = limits_profile.outputLimit

    for modifier in res.modifiers.values():
        # Clean up modifiers coming from the package that are not overridden
        # by the base limits profile.
        if time_is_overridden:
            modifier.time = None
        if memory_is_overridden:
            modifier.memory = None

    for language, modifier in limits_profile.modifiers.items():
        if modifier.time is not None:
            res.modifiers[language].time = modifier.time
        if modifier.timeMultiplier is not None:
            res.modifiers[language].timeMultiplier = modifier.timeMultiplier
        if modifier.memory is not None:
            res.modifiers[language].memory = modifier.memory
    return res


def _get_limits_from_profile(
    language: Optional[str],
    limits_profile: LimitsProfile,
    source_profile: Optional[str],
    verification: VerificationLevel,
    root: pathlib.Path,
) -> Limits:
    limits_profile = _expand_limits_profile(limits_profile, root=root)
    return Limits(
        time=limits_profile.timelimit_for_language(language),
        memory=limits_profile.memorylimit_for_language(language),
        output=limits_profile.outputLimit,
        isDoubleTL=verification.value >= VerificationLevel.FULL.value,
        profile=source_profile,
    )


def get_saved_limits_profile(
    profile: str = 'local', root: pathlib.Path = pathlib.Path()
) -> Optional[LimitsProfile]:
    limits_path = package.get_limits_file(profile, root=root)
    if not limits_path.exists():
        return None
    return utils.model_from_yaml(LimitsProfile, limits_path.read_text())


def get_package_limits_profile(root: pathlib.Path = pathlib.Path()) -> LimitsProfile:
    profile = LimitsProfile(inheritFromPackage=True)
    return _expand_limits_profile(profile, root=root)


def get_package_limits(
    verification: VerificationLevel = VerificationLevel.NONE,
    root: pathlib.Path = pathlib.Path(),
) -> Limits:
    return _get_limits_from_profile(
        language=None,
        limits_profile=get_package_limits_profile(root=root),
        source_profile=None,
        verification=verification,
        root=root,
    )


def get_limits_profile(
    profile: Optional[str] = None,
    fallback_to_package_profile: bool = True,
    root: pathlib.Path = pathlib.Path(),
) -> LimitsProfile:
    if profile is None:
        return get_package_limits_profile(root=root)
    saved_profile = get_saved_limits_profile(profile, root=root)
    if saved_profile is None:
        if fallback_to_package_profile:
            return get_package_limits_profile(root=root)
        console.console.print(
            f'[error]Limits profile [item]{profile}[/item] not found.[/error]'
        )
        raise typer.Exit(1)
    return _expand_limits_profile(saved_profile, root=root)


def get_limits(
    language: Optional[str] = None,
    profile: Optional[str] = None,
    fallback_to_package_profile: bool = True,
    verification: VerificationLevel = VerificationLevel.NONE,
    root: pathlib.Path = pathlib.Path(),
) -> Limits:
    source_profile = None
    limits_profile = LimitsProfile(inheritFromPackage=True)
    if profile is not None:
        specified_limits_profile = get_saved_limits_profile(profile, root=root)
        if specified_limits_profile is not None:
            limits_profile = specified_limits_profile
            source_profile = profile
        elif not fallback_to_package_profile:
            console.console.print(
                f'[error]Limits profile [item]{profile}[/item] not found.[/error]'
            )
            raise typer.Exit(1)

    res = _get_limits_from_profile(
        language, limits_profile, source_profile, verification, root=root
    )
    return res



import pathlib
from typing import Type

from textual.app import App, ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Footer, Header, OptionList

from rbx.box import remote
from rbx.box.ui.screens.differ import DifferScreen
from rbx.box.ui.screens.run_explorer import RunExplorerScreen
from rbx.box.ui.screens.test_explorer import TestExplorerScreen

SCREEN_OPTIONS = [
    ('Explore tests built by `rbx build`.', TestExplorerScreen),
    ('Explore results of a past `rbx run`.', RunExplorerScreen),
]


class rbxApp(App):
    TITLE = 'rbx'
    CSS_PATH = 'css/app.tcss'
    BINDINGS = [('q', 'quit', 'Quit')]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Center(id='main'):
            yield OptionList(*(opt[0] for opt in SCREEN_OPTIONS))

    def on_mount(self):
        self.query_one(OptionList).border_title = 'Select a flow'

    def on_option_list_option_selected(self, event: OptionList.OptionSelected):
        self.show_screen(SCREEN_OPTIONS[event.option_index][1])

    def show_screen(self, screen_cls: Type[Screen]):
        self.push_screen(screen_cls())


class rbxDifferApp(App):
    TITLE = 'rbx differ'
    CSS_PATH = 'css/app.tcss'
    BINDINGS = [('q', 'quit', 'Quit')]

    def __init__(self, path1: pathlib.Path, path2: pathlib.Path):
        super().__init__()
        self.path1 = path1
        self.path2 = path2

    def on_mount(self):
        self.push_screen(DifferScreen(self.path1, self.path2))


def start():
    app = rbxApp()
    app.run()


def start_differ(path1: pathlib.Path, path2: pathlib.Path):
    path1, path2 = remote.expand_files([path1, path2])

    app = rbxDifferApp(path1, path2)
    app.run()



from typing import List

from rbx.box import environment, package
from rbx.box.contest import contest_package
from rbx.box.contest.schema import ContestProblem


def match_problem(problems: str, contest_problem: ContestProblem) -> bool:
    short_name = contest_problem.short_name.lower()
    problems = problems.lower()
    if problems == '*':
        return True
    if '-' in problems:
        start, end = problems.split('-')
        return start <= short_name <= end
    problem_set = set(p.strip().lower() for p in problems.split(','))
    return short_name in problem_set


def get_problems_of_interest(problems: str) -> List[ContestProblem]:
    contest = contest_package.find_contest_package_or_die()
    problems_of_interest = []

    for p in contest.problems:
        if match_problem(problems, p):
            problems_of_interest.append(p)
    return problems_of_interest


def clear_all_caches():
    pkgs = [package, environment, contest_package]

    for pkg in pkgs:
        for fn in pkg.__dict__.values():
            if hasattr(fn, 'cache_clear'):
                fn.cache_clear()
